"""Metadata types for GQLDB Python driver."""

from dataclasses import dataclass, field
from typing import Dict, Optional

from .enums import GraphType

@dataclass
class GraphInfo:
    """Information about a graph."""
    name: str
    graph_type: GraphType
    node_count: int
    edge_count: int
    description: str


@dataclass
class TransactionInfo:
    """Information about a transaction."""
    transaction_id: int
    session_id: int
    graph_name: str
    read_only: bool
    created_at: int
    duration_ms: int
    internal_tx_id: str


@dataclass
class ASTCacheStats:
    """AST cache statistics."""
    hits: int
    misses: int
    evictions: int
    entries: int
    hit_rate: float


@dataclass
class PlanCacheStats:
    """Plan cache statistics."""
    size: int
    capacity: int
    hits: int
    misses: int
    hit_rate: float


@dataclass
class CacheStats:
    """Combined cache statistics."""
    ast_stats: Optional[ASTCacheStats] = None
    plan_stats: Optional[PlanCacheStats] = None


@dataclass
class Statistics:
    """Database statistics."""
    node_count: int
    edge_count: int
    label_counts: Dict[str, int] = field(default_factory=dict)
    edge_label_counts: Dict[str, int] = field(default_factory=dict)



"""Integration tests for as_* methods and printers."""

import pytest
import uuid

from gqldb.client import GqldbClient, QueryConfig
from gqldb.types import GraphType, NodeData, EdgeData
from gqldb.printers import (
    print_nodes,
    print_edges,
    print_paths,
    print_any,
    print_nodes_without_schema,
    print_edges_without_schema,
)
from tests.conftest import drop_test_graph


@pytest.mark.integration
class TestAsMethodsIntegration:
    """Integration tests for as_* methods."""

    @pytest.fixture
    def test_graph(self, connected_client: GqldbClient, request):
        """Create a temporary test graph."""
        graph_name = f"test_as_{request.node.name}_{uuid.uuid4().hex[:8]}"
        connected_client.create_graph(graph_name, GraphType.OPEN, "Test as methods")
        yield graph_name
        drop_test_graph(connected_client, graph_name)

    def test_as_nodes_with_real_data(self, connected_client: GqldbClient, test_graph: str):
        """Test as_nodes() extracts nodes from real query results."""
        # Start bulk import session
        session = connected_client.start_bulk_import(test_graph)

        try:
            # Insert test nodes
            nodes = [
                NodeData(labels=["Person"], properties={"name": "Alice", "age": 30}),
                NodeData(labels=["Person"], properties={"name": "Bob", "age": 25}),
            ]
            result = connected_client.insert_nodes(
                test_graph,
                nodes,
                bulk_import_session_id=session.session_id
            )
            assert result.success

        finally:
            connected_client.end_bulk_import(session.session_id)

        # Query and extract nodes - use RETURN n to get PropertyType.NODE typed values
        resp = connected_client.gql(
            "MATCH (n:Person) RETURN n",
            QueryConfig(graph_name=test_graph)
        )

        # Extract nodes from column "n"
        ar = resp.alias("n")
        extracted_nodes, schemas = ar.as_nodes()

        assert len(extracted_nodes) >= 2

        # Verify some data was extracted
        names = [n.properties.get("name") for n in extracted_nodes if n.properties.get("name")]
        assert "Alice" in names
        assert "Bob" in names

        # Test printer
        output = print_nodes(extracted_nodes, schemas)
        print(f"Printed nodes:\n{output}")
        assert "node(s)" in output

    def test_as_edges_with_real_data(self, connected_client: GqldbClient, test_graph: str):
        """Test as_edges() extracts edges from real query results."""
        # Start bulk import session
        session = connected_client.start_bulk_import(test_graph)

        try:
            # Insert nodes first
            nodes = [
                NodeData(labels=["EdgeTestPerson"], properties={"name": "Dave"}),
                NodeData(labels=["EdgeTestPerson"], properties={"name": "Eve"}),
            ]
            node_result = connected_client.insert_nodes(
                test_graph,
                nodes,
                bulk_import_session_id=session.session_id
            )
            assert node_result.success

            # Query to get node IDs
            query_config = QueryConfig(graph_name=test_graph)
            resp = connected_client.gql(
                "MATCH (n:EdgeTestPerson) RETURN id(n) AS node_id ORDER BY n.name",
                query_config
            )

            node_ids = []
            for i in range(resp.row_count):
                row = resp.rows[i]
                value = resp.get_by_name(row, "node_id")
                if value:
                    node_ids.append(value)

            if len(node_ids) >= 2:
                # Insert edge
                edges = [
                    EdgeData(
                        label="KNOWS_TEST",
                        from_node_id=node_ids[0],
                        to_node_id=node_ids[1],
                        properties={"since": 2024, "strength": 0.8}
                    )
                ]
                edge_result = connected_client.insert_edges(
                    test_graph,
                    edges,
                    bulk_import_session_id=session.session_id
                )
                assert edge_result.success

        finally:
            connected_client.end_bulk_import(session.session_id)

        # Query and extract edges - use RETURN e to get PropertyType.EDGE typed values
        resp = connected_client.gql(
            "MATCH (a)-[e:KNOWS_TEST]->(b) RETURN e",
            QueryConfig(graph_name=test_graph)
        )

        # Extract edges from column "e"
        ar = resp.alias("e")
        extracted_edges, schemas = ar.as_edges()

        assert len(extracted_edges) >= 1

        # Test printer
        output = print_edges(extracted_edges, schemas)
        print(f"Printed edges:\n{output}")
        assert "edge(s)" in output

    def test_as_table_with_real_data(self, connected_client: GqldbClient, test_graph: str):
        """Test as_table() converts query results to table."""
        # Start bulk import session
        session = connected_client.start_bulk_import(test_graph)

        try:
            # Insert test nodes
            nodes = [
                NodeData(labels=["TableTest"], properties={"val": 1}),
                NodeData(labels=["TableTest"], properties={"val": 2}),
            ]
            connected_client.insert_nodes(
                test_graph,
                nodes,
                bulk_import_session_id=session.session_id
            )

        finally:
            connected_client.end_bulk_import(session.session_id)

        # Query and get table
        resp = connected_client.gql(
            "MATCH (n:TableTest) RETURN n.val as value ORDER BY n.val",
            QueryConfig(graph_name=test_graph)
        )

        # Extract table from column "value"
        ar = resp.alias("value")
        table = ar.as_table()

        assert len(table.headers) >= 1
        print(f"Table headers: {[h.name for h in table.headers]}")
        print(f"Table rows: {len(table.rows)}")

    def test_as_attr_with_real_data(self, connected_client: GqldbClient, test_graph: str):
        """Test as_attr() extracts attribute values."""
        # Start bulk import session
        session = connected_client.start_bulk_import(test_graph)

        try:
            # Insert test nodes
            nodes = [
                NodeData(labels=["AttrTest"], properties={"name": "Alice"}),
                NodeData(labels=["AttrTest"], properties={"name": "Bob"}),
                NodeData(labels=["AttrTest"], properties={"name": "Charlie"}),
            ]
            connected_client.insert_nodes(
                test_graph,
                nodes,
                bulk_import_session_id=session.session_id
            )

        finally:
            connected_client.end_bulk_import(session.session_id)

        # Query and extract attribute
        resp = connected_client.gql(
            "MATCH (n:AttrTest) RETURN n.name as name",
            QueryConfig(graph_name=test_graph)
        )

        if len(resp) > 0:
            # Extract attribute from column "name"
            ar = resp.alias("name")
            attr = ar.as_attr()
            assert attr.name == "name"
            assert len(attr.values) >= 1
            print(f"Extracted names: {attr.values}")

    def test_as_attr_column_not_found(self, connected_client: GqldbClient, test_graph: str):
        """Test alias() throws for non-existent column."""
        resp = connected_client.gql("MATCH (n) RETURN n.name as name LIMIT 1", QueryConfig(graph_name=test_graph))

        with pytest.raises(KeyError, match="Column alias not found"):
            resp.alias("nonexistent")


@pytest.mark.integration
class TestPrintAnyIntegration:
    """Integration tests for print_any function."""

    @pytest.fixture
    def test_graph(self, connected_client: GqldbClient, request):
        """Create a temporary test graph."""
        graph_name = f"test_print_{request.node.name}_{uuid.uuid4().hex[:8]}"
        connected_client.create_graph(graph_name, GraphType.OPEN, "Test print_any")
        yield graph_name
        drop_test_graph(connected_client, graph_name)

    def test_auto_detect_nodes(self, connected_client: GqldbClient, test_graph: str):
        """Test print_any() auto-detects node response."""
        # Start bulk import session
        session = connected_client.start_bulk_import(test_graph)

        try:
            nodes = [NodeData(labels=["PrintAnyTest"], properties={"val": 1})]
            connected_client.insert_nodes(
                test_graph,
                nodes,
                bulk_import_session_id=session.session_id
            )

        finally:
            connected_client.end_bulk_import(session.session_id)

        # Use RETURN n to get PropertyType.NODE typed values
        resp = connected_client.gql(
            "MATCH (n:PrintAnyTest) RETURN n",
            QueryConfig(graph_name=test_graph)
        )

        output = print_any(resp)
        assert output is not None
        print(f"Auto-detected (nodes):\n{output}")

    def test_auto_detect_edges(self, connected_client: GqldbClient, test_graph: str):
        """Test print_any() auto-detects edge response."""
        # Start bulk import session
        session = connected_client.start_bulk_import(test_graph)

        try:
            nodes = [
                NodeData(labels=["PANode"], properties={"name": "A"}),
                NodeData(labels=["PANode"], properties={"name": "B"}),
            ]
            node_result = connected_client.insert_nodes(
                test_graph,
                nodes,
                bulk_import_session_id=session.session_id
            )

            # Query to get node IDs
            query_config = QueryConfig(graph_name=test_graph)
            resp = connected_client.gql(
                "MATCH (n:PANode) RETURN id(n) AS node_id ORDER BY n.name",
                query_config
            )

            node_ids = []
            for i in range(resp.row_count):
                row = resp.rows[i]
                value = resp.get_by_name(row, "node_id")
                if value:
                    node_ids.append(value)

            if len(node_ids) >= 2:
                edges = [
                    EdgeData(
                        label="PA_LINK",
                        from_node_id=node_ids[0],
                        to_node_id=node_ids[1],
                        properties={}
                    )
                ]
                connected_client.insert_edges(
                    test_graph,
                    edges,
                    bulk_import_session_id=session.session_id
                )

        finally:
            connected_client.end_bulk_import(session.session_id)

        # Use RETURN e to get PropertyType.EDGE typed values
        resp = connected_client.gql(
            "MATCH (a)-[e:PA_LINK]->(b) RETURN e",
            QueryConfig(graph_name=test_graph)
        )

        output = print_any(resp)
        assert output is not None
        print(f"Auto-detected (edges):\n{output}")

    def test_fallback_to_table(self, connected_client: GqldbClient, test_graph: str):
        """Test print_any() falls back to table format."""
        resp = connected_client.gql("RETURN 1 + 1 as result, 'hello' as greeting", QueryConfig(graph_name=test_graph))

        output = print_any(resp)
        assert output is not None
        print(f"Auto-detected (table):\n{output}")

    def test_empty_response(self, connected_client: GqldbClient, test_graph: str):
        """Test print_any() handles empty response."""
        resp = connected_client.gql(
            "MATCH (n:NonExistentLabel12345) RETURN n LIMIT 1",
            QueryConfig(graph_name=test_graph)
        )

        output = print_any(resp)
        assert "No data" in output



"""Specialized data types for GQLDB Python driver."""

from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import List


@dataclass
class Point:
    """2D geographic point."""
    latitude: float
    longitude: float


@dataclass
class Point3D:
    """3D point."""
    x: float
    y: float
    z: float


@dataclass
class GqldbDecimal:
    """High-precision decimal number."""
    value: str


@dataclass
class GqldbLocalDateTime:
    """Local date-time without timezone, stored as components."""
    year: int
    month: int  # 1-12
    day: int  # 1-31
    hour: int  # 0-23
    minute: int  # 0-59
    second: int  # 0-59
    nanosecond: int = 0  # 0-999999999

    def to_datetime(self) -> datetime:
        """Convert to a datetime (no timezone).

        The returned datetime uses the stored components directly.
        Microsecond precision only (nanoseconds are truncated to microseconds).
        """
        return datetime(
            self.year, self.month, self.day,
            self.hour, self.minute, self.second,
            self.nanosecond // 1000,
        )

    def __str__(self) -> str:
        """Return ISO 8601 format without timezone (e.g., '2024-06-15T14:30:00')."""
        dt = self.to_datetime()
        if dt.microsecond:
            return dt.strftime("%Y-%m-%dT%H:%M:%S.") + f"{dt.microsecond:06d}"
        return dt.strftime("%Y-%m-%dT%H:%M:%S")


@dataclass
class GqldbZonedDateTime:
    """Date-time with timezone offset, stored as components."""
    year: int
    month: int  # 1-12
    day: int  # 1-31
    hour: int  # 0-23
    minute: int  # 0-59
    second: int  # 0-59
    nanosecond: int = 0  # 0-999999999
    offset_minutes: int = 0  # UTC offset in minutes (signed)

    def to_datetime(self) -> datetime:
        """Convert to a timezone-aware datetime.

        The returned datetime uses the stored components with the timezone offset applied.
        Microsecond precision only (nanoseconds are truncated to microseconds).
        """
        from datetime import timedelta
        tz = timezone(timedelta(minutes=self.offset_minutes))
        return datetime(
            self.year, self.month, self.day,
            self.hour, self.minute, self.second,
            self.nanosecond // 1000,
            tzinfo=tz,
        )


@dataclass
class GqldbDate:
    """Date without time."""
    year: int
    month: int  # 1-12
    day: int  # 1-31


@dataclass
class GqldbLocalTime:
    """Local time without timezone."""
    hour: int
    minute: int
    second: int
    nanosecond: int


@dataclass
class GqldbZonedTime:
    """Time with timezone offset."""
    hour: int
    minute: int
    second: int
    nanosecond: int
    offset_minutes: int


@dataclass
class YearToMonth:
    """Year-month interval."""
    months: int  # total months (can be negative)


@dataclass
class DayToSecond:
    """Day-second interval."""
    seconds: int
    nanoseconds: int


@dataclass
class Vector:
    """Float32 vector for embeddings."""
    values: List[float] = field(default_factory=list)


@dataclass
class GqldbTable:
    """Table data with columns and rows."""
    columns: List[str] = field(default_factory=list)
    rows: List[List] = field(default_factory=list)
    column_types: List[int] = field(default_factory=list)


@dataclass
class ComputeTopologyResult:
    """Result of wait_for_compute_topology operation."""
    ready: bool
    message: str



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import gqldb_pb2 as gqldb__pb2

GRPC_GENERATED_VERSION = '1.76.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in gqldb_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class SessionServiceStub(object):
    """=============================================================================
    Session Service
    =============================================================================

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Login = channel.unary_unary(
                '/gqldb.SessionService/Login',
                request_serializer=gqldb__pb2.LoginRequest.SerializeToString,
                response_deserializer=gqldb__pb2.LoginResponse.FromString,
                _registered_method=True)
        self.Logout = channel.unary_unary(
                '/gqldb.SessionService/Logout',
                request_serializer=gqldb__pb2.LogoutRequest.SerializeToString,
                response_deserializer=gqldb__pb2.LogoutResponse.FromString,
                _registered_method=True)
        self.Ping = channel.unary_unary(
                '/gqldb.SessionService/Ping',
                request_serializer=gqldb__pb2.PingRequest.SerializeToString,
                response_deserializer=gqldb__pb2.PingResponse.FromString,
                _registered_method=True)


class SessionServiceServicer(object):
    """=============================================================================
    Session Service
    =============================================================================

    """

    def Login(self, request, context):
        """Login authenticates a user and creates a session.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Logout(self, request, context):
        """Logout terminates a session.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Ping(self, request, context):
        """Ping keeps a session alive and checks connectivity.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SessionServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Login': grpc.unary_unary_rpc_method_handler(
                    servicer.Login,
                    request_deserializer=gqldb__pb2.LoginRequest.FromString,
                    response_serializer=gqldb__pb2.LoginResponse.SerializeToString,
            ),
            'Logout': grpc.unary_unary_rpc_method_handler(
                    servicer.Logout,
                    request_deserializer=gqldb__pb2.LogoutRequest.FromString,
                    response_serializer=gqldb__pb2.LogoutResponse.SerializeToString,
            ),
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=gqldb__pb2.PingRequest.FromString,
                    response_serializer=gqldb__pb2.PingResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gqldb.SessionService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gqldb.SessionService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class SessionService(object):
    """=============================================================================
    Session Service
    =============================================================================

    """

    @staticmethod
    def Login(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.SessionService/Login',
            gqldb__pb2.LoginRequest.SerializeToString,
            gqldb__pb2.LoginResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Logout(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.SessionService/Logout',
            gqldb__pb2.LogoutRequest.SerializeToString,
            gqldb__pb2.LogoutResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.SessionService/Ping',
            gqldb__pb2.PingRequest.SerializeToString,
            gqldb__pb2.PingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class QueryServiceStub(object):
    """=============================================================================
    Query Service
    =============================================================================

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Gql = channel.unary_unary(
                '/gqldb.QueryService/Gql',
                request_serializer=gqldb__pb2.GqlRequest.SerializeToString,
                response_deserializer=gqldb__pb2.GqlResponse.FromString,
                _registered_method=True)
        self.GqlStream = channel.unary_stream(
                '/gqldb.QueryService/GqlStream',
                request_serializer=gqldb__pb2.GqlRequest.SerializeToString,
                response_deserializer=gqldb__pb2.GqlResponse.FromString,
                _registered_method=True)
        self.GqlStatement = channel.unary_unary(
                '/gqldb.QueryService/GqlStatement',
                request_serializer=gqldb__pb2.GqlRequest.SerializeToString,
                response_deserializer=gqldb__pb2.StatementResult.FromString,
                _registered_method=True)
        self.Explain = channel.unary_unary(
                '/gqldb.QueryService/Explain',
                request_serializer=gqldb__pb2.GqlRequest.SerializeToString,
                response_deserializer=gqldb__pb2.ExplainResponse.FromString,
                _registered_method=True)
        self.Profile = channel.unary_unary(
                '/gqldb.QueryService/Profile',
                request_serializer=gqldb__pb2.GqlRequest.SerializeToString,
                response_deserializer=gqldb__pb2.ProfileResponse.FromString,
                _registered_method=True)


class QueryServiceServicer(object):
    """=============================================================================
    Query Service
    =============================================================================

    """

    def Gql(self, request, context):
        """Gql executes a GQL query and returns results.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GqlStream(self, request, context):
        """GqlStream executes a GQL query and streams results.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GqlStatement(self, request, context):
        """GqlStatement executes a GQL statement (DDL/DML) and returns status.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Explain(self, request, context):
        """Explain returns the execution plan for a query.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Profile(self, request, context):
        """Profile executes a query with profiling and returns statistics.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_QueryServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Gql': grpc.unary_unary_rpc_method_handler(
                    servicer.Gql,
                    request_deserializer=gqldb__pb2.GqlRequest.FromString,
                    response_serializer=gqldb__pb2.GqlResponse.SerializeToString,
            ),
            'GqlStream': grpc.unary_stream_rpc_method_handler(
                    servicer.GqlStream,
                    request_deserializer=gqldb__pb2.GqlRequest.FromString,
                    response_serializer=gqldb__pb2.GqlResponse.SerializeToString,
            ),
            'GqlStatement': grpc.unary_unary_rpc_method_handler(
                    servicer.GqlStatement,
                    request_deserializer=gqldb__pb2.GqlRequest.FromString,
                    response_serializer=gqldb__pb2.StatementResult.SerializeToString,
            ),
            'Explain': grpc.unary_unary_rpc_method_handler(
                    servicer.Explain,
                    request_deserializer=gqldb__pb2.GqlRequest.FromString,
                    response_serializer=gqldb__pb2.ExplainResponse.SerializeToString,
            ),
            'Profile': grpc.unary_unary_rpc_method_handler(
                    servicer.Profile,
                    request_deserializer=gqldb__pb2.GqlRequest.FromString,
                    response_serializer=gqldb__pb2.ProfileResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gqldb.QueryService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gqldb.QueryService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class QueryService(object):
    """=============================================================================
    Query Service
    =============================================================================

    """

    @staticmethod
    def Gql(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.QueryService/Gql',
            gqldb__pb2.GqlRequest.SerializeToString,
            gqldb__pb2.GqlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GqlStream(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/gqldb.QueryService/GqlStream',
            gqldb__pb2.GqlRequest.SerializeToString,
            gqldb__pb2.GqlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GqlStatement(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.QueryService/GqlStatement',
            gqldb__pb2.GqlRequest.SerializeToString,
            gqldb__pb2.StatementResult.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Explain(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.QueryService/Explain',
            gqldb__pb2.GqlRequest.SerializeToString,
            gqldb__pb2.ExplainResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Profile(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.QueryService/Profile',
            gqldb__pb2.GqlRequest.SerializeToString,
            gqldb__pb2.ProfileResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class DataServiceStub(object):
    """=============================================================================
    Data Service
    =============================================================================

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.InsertNodes = channel.unary_unary(
                '/gqldb.DataService/InsertNodes',
                request_serializer=gqldb__pb2.InsertNodesRequest.SerializeToString,
                response_deserializer=gqldb__pb2.InsertNodesResponse.FromString,
                _registered_method=True)
        self.InsertEdges = channel.unary_unary(
                '/gqldb.DataService/InsertEdges',
                request_serializer=gqldb__pb2.InsertEdgesRequest.SerializeToString,
                response_deserializer=gqldb__pb2.InsertEdgesResponse.FromString,
                _registered_method=True)
        self.DeleteNodes = channel.unary_unary(
                '/gqldb.DataService/DeleteNodes',
                request_serializer=gqldb__pb2.DeleteNodesRequest.SerializeToString,
                response_deserializer=gqldb__pb2.DeleteNodesResponse.FromString,
                _registered_method=True)
        self.DeleteEdges = channel.unary_unary(
                '/gqldb.DataService/DeleteEdges',
                request_serializer=gqldb__pb2.DeleteEdgesRequest.SerializeToString,
                response_deserializer=gqldb__pb2.DeleteEdgesResponse.FromString,
                _registered_method=True)
        self.ExportNodes = channel.unary_stream(
                '/gqldb.DataService/ExportNodes',
                request_serializer=gqldb__pb2.ExportNodesRequest.SerializeToString,
                response_deserializer=gqldb__pb2.ExportNodesResponse.FromString,
                _registered_method=True)
        self.ExportEdges = channel.unary_stream(
                '/gqldb.DataService/ExportEdges',
                request_serializer=gqldb__pb2.ExportEdgesRequest.SerializeToString,
                response_deserializer=gqldb__pb2.ExportEdgesResponse.FromString,
                _registered_method=True)


class DataServiceServicer(object):
    """=============================================================================
    Data Service
    =============================================================================

    """

    def InsertNodes(self, request, context):
        """InsertNodes inserts multiple nodes into a graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InsertEdges(self, request, context):
        """InsertEdges inserts multiple edges into a graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteNodes(self, request, context):
        """DeleteNodes deletes nodes from a graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteEdges(self, request, context):
        """DeleteEdges deletes edges from a graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExportNodes(self, request, context):
        """ExportNodes streams nodes from a graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExportEdges(self, request, context):
        """ExportEdges streams edges from a graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DataServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'InsertNodes': grpc.unary_unary_rpc_method_handler(
                    servicer.InsertNodes,
                    request_deserializer=gqldb__pb2.InsertNodesRequest.FromString,
                    response_serializer=gqldb__pb2.InsertNodesResponse.SerializeToString,
            ),
            'InsertEdges': grpc.unary_unary_rpc_method_handler(
                    servicer.InsertEdges,
                    request_deserializer=gqldb__pb2.InsertEdgesRequest.FromString,
                    response_serializer=gqldb__pb2.InsertEdgesResponse.SerializeToString,
            ),
            'DeleteNodes': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteNodes,
                    request_deserializer=gqldb__pb2.DeleteNodesRequest.FromString,
                    response_serializer=gqldb__pb2.DeleteNodesResponse.SerializeToString,
            ),
            'DeleteEdges': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteEdges,
                    request_deserializer=gqldb__pb2.DeleteEdgesRequest.FromString,
                    response_serializer=gqldb__pb2.DeleteEdgesResponse.SerializeToString,
            ),
            'ExportNodes': grpc.unary_stream_rpc_method_handler(
                    servicer.ExportNodes,
                    request_deserializer=gqldb__pb2.ExportNodesRequest.FromString,
                    response_serializer=gqldb__pb2.ExportNodesResponse.SerializeToString,
            ),
            'ExportEdges': grpc.unary_stream_rpc_method_handler(
                    servicer.ExportEdges,
                    request_deserializer=gqldb__pb2.ExportEdgesRequest.FromString,
                    response_serializer=gqldb__pb2.ExportEdgesResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gqldb.DataService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gqldb.DataService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class DataService(object):
    """=============================================================================
    Data Service
    =============================================================================

    """

    @staticmethod
    def InsertNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.DataService/InsertNodes',
            gqldb__pb2.InsertNodesRequest.SerializeToString,
            gqldb__pb2.InsertNodesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def InsertEdges(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.DataService/InsertEdges',
            gqldb__pb2.InsertEdgesRequest.SerializeToString,
            gqldb__pb2.InsertEdgesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.DataService/DeleteNodes',
            gqldb__pb2.DeleteNodesRequest.SerializeToString,
            gqldb__pb2.DeleteNodesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteEdges(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.DataService/DeleteEdges',
            gqldb__pb2.DeleteEdgesRequest.SerializeToString,
            gqldb__pb2.DeleteEdgesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ExportNodes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/gqldb.DataService/ExportNodes',
            gqldb__pb2.ExportNodesRequest.SerializeToString,
            gqldb__pb2.ExportNodesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ExportEdges(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/gqldb.DataService/ExportEdges',
            gqldb__pb2.ExportEdgesRequest.SerializeToString,
            gqldb__pb2.ExportEdgesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class GraphServiceStub(object):
    """=============================================================================
    Graph Service
    =============================================================================

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateGraph = channel.unary_unary(
                '/gqldb.GraphService/CreateGraph',
                request_serializer=gqldb__pb2.CreateGraphRequest.SerializeToString,
                response_deserializer=gqldb__pb2.CreateGraphResponse.FromString,
                _registered_method=True)
        self.DropGraph = channel.unary_unary(
                '/gqldb.GraphService/DropGraph',
                request_serializer=gqldb__pb2.DropGraphRequest.SerializeToString,
                response_deserializer=gqldb__pb2.DropGraphResponse.FromString,
                _registered_method=True)
        self.UseGraph = channel.unary_unary(
                '/gqldb.GraphService/UseGraph',
                request_serializer=gqldb__pb2.UseGraphRequest.SerializeToString,
                response_deserializer=gqldb__pb2.UseGraphResponse.FromString,
                _registered_method=True)
        self.ListGraphs = channel.unary_unary(
                '/gqldb.GraphService/ListGraphs',
                request_serializer=gqldb__pb2.ListGraphsRequest.SerializeToString,
                response_deserializer=gqldb__pb2.ListGraphsResponse.FromString,
                _registered_method=True)
        self.GetGraphInfo = channel.unary_unary(
                '/gqldb.GraphService/GetGraphInfo',
                request_serializer=gqldb__pb2.GetGraphInfoRequest.SerializeToString,
                response_deserializer=gqldb__pb2.GetGraphInfoResponse.FromString,
                _registered_method=True)


class GraphServiceServicer(object):
    """=============================================================================
    Graph Service
    =============================================================================

    """

    def CreateGraph(self, request, context):
        """CreateGraph creates a new graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DropGraph(self, request, context):
        """DropGraph deletes a graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UseGraph(self, request, context):
        """UseGraph sets the current graph for a session.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListGraphs(self, request, context):
        """ListGraphs returns all available graphs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetGraphInfo(self, request, context):
        """GetGraphInfo returns detailed information about a graph.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GraphServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateGraph': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateGraph,
                    request_deserializer=gqldb__pb2.CreateGraphRequest.FromString,
                    response_serializer=gqldb__pb2.CreateGraphResponse.SerializeToString,
            ),
            'DropGraph': grpc.unary_unary_rpc_method_handler(
                    servicer.DropGraph,
                    request_deserializer=gqldb__pb2.DropGraphRequest.FromString,
                    response_serializer=gqldb__pb2.DropGraphResponse.SerializeToString,
            ),
            'UseGraph': grpc.unary_unary_rpc_method_handler(
                    servicer.UseGraph,
                    request_deserializer=gqldb__pb2.UseGraphRequest.FromString,
                    response_serializer=gqldb__pb2.UseGraphResponse.SerializeToString,
            ),
            'ListGraphs': grpc.unary_unary_rpc_method_handler(
                    servicer.ListGraphs,
                    request_deserializer=gqldb__pb2.ListGraphsRequest.FromString,
                    response_serializer=gqldb__pb2.ListGraphsResponse.SerializeToString,
            ),
            'GetGraphInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetGraphInfo,
                    request_deserializer=gqldb__pb2.GetGraphInfoRequest.FromString,
                    response_serializer=gqldb__pb2.GetGraphInfoResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gqldb.GraphService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gqldb.GraphService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class GraphService(object):
    """=============================================================================
    Graph Service
    =============================================================================

    """

    @staticmethod
    def CreateGraph(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.GraphService/CreateGraph',
            gqldb__pb2.CreateGraphRequest.SerializeToString,
            gqldb__pb2.CreateGraphResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DropGraph(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.GraphService/DropGraph',
            gqldb__pb2.DropGraphRequest.SerializeToString,
            gqldb__pb2.DropGraphResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UseGraph(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.GraphService/UseGraph',
            gqldb__pb2.UseGraphRequest.SerializeToString,
            gqldb__pb2.UseGraphResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListGraphs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.GraphService/ListGraphs',
            gqldb__pb2.ListGraphsRequest.SerializeToString,
            gqldb__pb2.ListGraphsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetGraphInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.GraphService/GetGraphInfo',
            gqldb__pb2.GetGraphInfoRequest.SerializeToString,
            gqldb__pb2.GetGraphInfoResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class TransactionServiceStub(object):
    """=============================================================================
    Transaction Service
    =============================================================================

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Begin = channel.unary_unary(
                '/gqldb.TransactionService/Begin',
                request_serializer=gqldb__pb2.BeginRequest.SerializeToString,
                response_deserializer=gqldb__pb2.BeginResponse.FromString,
                _registered_method=True)
        self.Commit = channel.unary_unary(
                '/gqldb.TransactionService/Commit',
                request_serializer=gqldb__pb2.CommitRequest.SerializeToString,
                response_deserializer=gqldb__pb2.CommitResponse.FromString,
                _registered_method=True)
        self.Rollback = channel.unary_unary(
                '/gqldb.TransactionService/Rollback',
                request_serializer=gqldb__pb2.RollbackRequest.SerializeToString,
                response_deserializer=gqldb__pb2.RollbackResponse.FromString,
                _registered_method=True)
        self.ListTransactions = channel.unary_unary(
                '/gqldb.TransactionService/ListTransactions',
                request_serializer=gqldb__pb2.ListTransactionsRequest.SerializeToString,
                response_deserializer=gqldb__pb2.ListTransactionsResponse.FromString,
                _registered_method=True)


class TransactionServiceServicer(object):
    """=============================================================================
    Transaction Service
    =============================================================================

    """

    def Begin(self, request, context):
        """Begin starts a new transaction.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Commit(self, request, context):
        """Commit commits a transaction.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Rollback(self, request, context):
        """Rollback aborts a transaction.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListTransactions(self, request, context):
        """ListTransactions returns active transactions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TransactionServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Begin': grpc.unary_unary_rpc_method_handler(
                    servicer.Begin,
                    request_deserializer=gqldb__pb2.BeginRequest.FromString,
                    response_serializer=gqldb__pb2.BeginResponse.SerializeToString,
            ),
            'Commit': grpc.unary_unary_rpc_method_handler(
                    servicer.Commit,
                    request_deserializer=gqldb__pb2.CommitRequest.FromString,
                    response_serializer=gqldb__pb2.CommitResponse.SerializeToString,
            ),
            'Rollback': grpc.unary_unary_rpc_method_handler(
                    servicer.Rollback,
                    request_deserializer=gqldb__pb2.RollbackRequest.FromString,
                    response_serializer=gqldb__pb2.RollbackResponse.SerializeToString,
            ),
            'ListTransactions': grpc.unary_unary_rpc_method_handler(
                    servicer.ListTransactions,
                    request_deserializer=gqldb__pb2.ListTransactionsRequest.FromString,
                    response_serializer=gqldb__pb2.ListTransactionsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gqldb.TransactionService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gqldb.TransactionService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class TransactionService(object):
    """=============================================================================
    Transaction Service
    =============================================================================

    """

    @staticmethod
    def Begin(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.TransactionService/Begin',
            gqldb__pb2.BeginRequest.SerializeToString,
            gqldb__pb2.BeginResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Commit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.TransactionService/Commit',
            gqldb__pb2.CommitRequest.SerializeToString,
            gqldb__pb2.CommitResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Rollback(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.TransactionService/Rollback',
            gqldb__pb2.RollbackRequest.SerializeToString,
            gqldb__pb2.RollbackResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListTransactions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.TransactionService/ListTransactions',
            gqldb__pb2.ListTransactionsRequest.SerializeToString,
            gqldb__pb2.ListTransactionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class HealthStub(object):
    """=============================================================================
    Health Service (gRPC Health Checking Protocol)
    =============================================================================

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Check = channel.unary_unary(
                '/gqldb.Health/Check',
                request_serializer=gqldb__pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=gqldb__pb2.HealthCheckResponse.FromString,
                _registered_method=True)
        self.Watch = channel.unary_stream(
                '/gqldb.Health/Watch',
                request_serializer=gqldb__pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=gqldb__pb2.HealthCheckResponse.FromString,
                _registered_method=True)


class HealthServicer(object):
    """=============================================================================
    Health Service (gRPC Health Checking Protocol)
    =============================================================================

    """

    def Check(self, request, context):
        """Check returns the serving status of the service.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Watch(self, request, context):
        """Watch streams serving status changes.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HealthServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Check': grpc.unary_unary_rpc_method_handler(
                    servicer.Check,
                    request_deserializer=gqldb__pb2.HealthCheckRequest.FromString,
                    response_serializer=gqldb__pb2.HealthCheckResponse.SerializeToString,
            ),
            'Watch': grpc.unary_stream_rpc_method_handler(
                    servicer.Watch,
                    request_deserializer=gqldb__pb2.HealthCheckRequest.FromString,
                    response_serializer=gqldb__pb2.HealthCheckResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gqldb.Health', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gqldb.Health', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Health(object):
    """=============================================================================
    Health Service (gRPC Health Checking Protocol)
    =============================================================================

    """

    @staticmethod
    def Check(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.Health/Check',
            gqldb__pb2.HealthCheckRequest.SerializeToString,
            gqldb__pb2.HealthCheckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Watch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/gqldb.Health/Watch',
            gqldb__pb2.HealthCheckRequest.SerializeToString,
            gqldb__pb2.HealthCheckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class AdminServiceStub(object):
    """=============================================================================
    Admin Service
    =============================================================================

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.WarmupParser = channel.unary_unary(
                '/gqldb.AdminService/WarmupParser',
                request_serializer=gqldb__pb2.WarmupParserRequest.SerializeToString,
                response_deserializer=gqldb__pb2.WarmupParserResponse.FromString,
                _registered_method=True)
        self.GetCacheStats = channel.unary_unary(
                '/gqldb.AdminService/GetCacheStats',
                request_serializer=gqldb__pb2.GetCacheStatsRequest.SerializeToString,
                response_deserializer=gqldb__pb2.CacheStatsResponse.FromString,
                _registered_method=True)
        self.ClearCache = channel.unary_unary(
                '/gqldb.AdminService/ClearCache',
                request_serializer=gqldb__pb2.ClearCacheRequest.SerializeToString,
                response_deserializer=gqldb__pb2.ClearCacheResponse.FromString,
                _registered_method=True)
        self.GetStatistics = channel.unary_unary(
                '/gqldb.AdminService/GetStatistics',
                request_serializer=gqldb__pb2.GetStatisticsRequest.SerializeToString,
                response_deserializer=gqldb__pb2.GetStatisticsResponse.FromString,
                _registered_method=True)
        self.InvalidatePermissionCache = channel.unary_unary(
                '/gqldb.AdminService/InvalidatePermissionCache',
                request_serializer=gqldb__pb2.InvalidatePermissionCacheRequest.SerializeToString,
                response_deserializer=gqldb__pb2.InvalidatePermissionCacheResponse.FromString,
                _registered_method=True)
        self.Compact = channel.unary_unary(
                '/gqldb.AdminService/Compact',
                request_serializer=gqldb__pb2.CompactRequest.SerializeToString,
                response_deserializer=gqldb__pb2.CompactResponse.FromString,
                _registered_method=True)


class AdminServiceServicer(object):
    """=============================================================================
    Admin Service
    =============================================================================

    """

    def WarmupParser(self, request, context):
        """WarmupParser pre-allocates parser instances for better performance.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCacheStats(self, request, context):
        """GetCacheStats returns cache statistics.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ClearCache(self, request, context):
        """ClearCache clears specified caches.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStatistics(self, request, context):
        """GetStatistics returns database statistics.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InvalidatePermissionCache(self, request, context):
        """InvalidatePermissionCache invalidates RBAC permission cache.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Compact(self, request, context):
        """Compact triggers manual compaction of the database storage.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AdminServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'WarmupParser': grpc.unary_unary_rpc_method_handler(
                    servicer.WarmupParser,
                    request_deserializer=gqldb__pb2.WarmupParserRequest.FromString,
                    response_serializer=gqldb__pb2.WarmupParserResponse.SerializeToString,
            ),
            'GetCacheStats': grpc.unary_unary_rpc_method_handler(
                    servicer.GetCacheStats,
                    request_deserializer=gqldb__pb2.GetCacheStatsRequest.FromString,
                    response_serializer=gqldb__pb2.CacheStatsResponse.SerializeToString,
            ),
            'ClearCache': grpc.unary_unary_rpc_method_handler(
                    servicer.ClearCache,
                    request_deserializer=gqldb__pb2.ClearCacheRequest.FromString,
                    response_serializer=gqldb__pb2.ClearCacheResponse.SerializeToString,
            ),
            'GetStatistics': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStatistics,
                    request_deserializer=gqldb__pb2.GetStatisticsRequest.FromString,
                    response_serializer=gqldb__pb2.GetStatisticsResponse.SerializeToString,
            ),
            'InvalidatePermissionCache': grpc.unary_unary_rpc_method_handler(
                    servicer.InvalidatePermissionCache,
                    request_deserializer=gqldb__pb2.InvalidatePermissionCacheRequest.FromString,
                    response_serializer=gqldb__pb2.InvalidatePermissionCacheResponse.SerializeToString,
            ),
            'Compact': grpc.unary_unary_rpc_method_handler(
                    servicer.Compact,
                    request_deserializer=gqldb__pb2.CompactRequest.FromString,
                    response_serializer=gqldb__pb2.CompactResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gqldb.AdminService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gqldb.AdminService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class AdminService(object):
    """=============================================================================
    Admin Service
    =============================================================================

    """

    @staticmethod
    def WarmupParser(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.AdminService/WarmupParser',
            gqldb__pb2.WarmupParserRequest.SerializeToString,
            gqldb__pb2.WarmupParserResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetCacheStats(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.AdminService/GetCacheStats',
            gqldb__pb2.GetCacheStatsRequest.SerializeToString,
            gqldb__pb2.CacheStatsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ClearCache(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.AdminService/ClearCache',
            gqldb__pb2.ClearCacheRequest.SerializeToString,
            gqldb__pb2.ClearCacheResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStatistics(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.AdminService/GetStatistics',
            gqldb__pb2.GetStatisticsRequest.SerializeToString,
            gqldb__pb2.GetStatisticsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def InvalidatePermissionCache(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.AdminService/InvalidatePermissionCache',
            gqldb__pb2.InvalidatePermissionCacheRequest.SerializeToString,
            gqldb__pb2.InvalidatePermissionCacheResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Compact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.AdminService/Compact',
            gqldb__pb2.CompactRequest.SerializeToString,
            gqldb__pb2.CompactResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class BulkImportServiceStub(object):
    """=============================================================================
    Bulk Import Service
    =============================================================================

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StartBulkImport = channel.unary_unary(
                '/gqldb.BulkImportService/StartBulkImport',
                request_serializer=gqldb__pb2.StartBulkImportRequest.SerializeToString,
                response_deserializer=gqldb__pb2.StartBulkImportResponse.FromString,
                _registered_method=True)
        self.Checkpoint = channel.unary_unary(
                '/gqldb.BulkImportService/Checkpoint',
                request_serializer=gqldb__pb2.CheckpointRequest.SerializeToString,
                response_deserializer=gqldb__pb2.CheckpointResponse.FromString,
                _registered_method=True)
        self.EndBulkImport = channel.unary_unary(
                '/gqldb.BulkImportService/EndBulkImport',
                request_serializer=gqldb__pb2.EndBulkImportRequest.SerializeToString,
                response_deserializer=gqldb__pb2.EndBulkImportResponse.FromString,
                _registered_method=True)
        self.AbortBulkImport = channel.unary_unary(
                '/gqldb.BulkImportService/AbortBulkImport',
                request_serializer=gqldb__pb2.AbortBulkImportRequest.SerializeToString,
                response_deserializer=gqldb__pb2.AbortBulkImportResponse.FromString,
                _registered_method=True)
        self.GetBulkImportStatus = channel.unary_unary(
                '/gqldb.BulkImportService/GetBulkImportStatus',
                request_serializer=gqldb__pb2.GetBulkImportStatusRequest.SerializeToString,
                response_deserializer=gqldb__pb2.GetBulkImportStatusResponse.FromString,
                _registered_method=True)


class BulkImportServiceServicer(object):
    """=============================================================================
    Bulk Import Service
    =============================================================================

    """

    def StartBulkImport(self, request, context):
        """StartBulkImport starts a bulk import session for optimized high-throughput inserts.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Checkpoint(self, request, context):
        """Checkpoint flushes all accumulated data to disk for durability.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EndBulkImport(self, request, context):
        """EndBulkImport ends the session with a final checkpoint.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AbortBulkImport(self, request, context):
        """AbortBulkImport cancels the session without final sync.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBulkImportStatus(self, request, context):
        """GetBulkImportStatus returns the current status of a bulk import session.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_BulkImportServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StartBulkImport': grpc.unary_unary_rpc_method_handler(
                    servicer.StartBulkImport,
                    request_deserializer=gqldb__pb2.StartBulkImportRequest.FromString,
                    response_serializer=gqldb__pb2.StartBulkImportResponse.SerializeToString,
            ),
            'Checkpoint': grpc.unary_unary_rpc_method_handler(
                    servicer.Checkpoint,
                    request_deserializer=gqldb__pb2.CheckpointRequest.FromString,
                    response_serializer=gqldb__pb2.CheckpointResponse.SerializeToString,
            ),
            'EndBulkImport': grpc.unary_unary_rpc_method_handler(
                    servicer.EndBulkImport,
                    request_deserializer=gqldb__pb2.EndBulkImportRequest.FromString,
                    response_serializer=gqldb__pb2.EndBulkImportResponse.SerializeToString,
            ),
            'AbortBulkImport': grpc.unary_unary_rpc_method_handler(
                    servicer.AbortBulkImport,
                    request_deserializer=gqldb__pb2.AbortBulkImportRequest.FromString,
                    response_serializer=gqldb__pb2.AbortBulkImportResponse.SerializeToString,
            ),
            'GetBulkImportStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBulkImportStatus,
                    request_deserializer=gqldb__pb2.GetBulkImportStatusRequest.FromString,
                    response_serializer=gqldb__pb2.GetBulkImportStatusResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gqldb.BulkImportService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('gqldb.BulkImportService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class BulkImportService(object):
    """=============================================================================
    Bulk Import Service
    =============================================================================

    """

    @staticmethod
    def StartBulkImport(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.BulkImportService/StartBulkImport',
            gqldb__pb2.StartBulkImportRequest.SerializeToString,
            gqldb__pb2.StartBulkImportResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Checkpoint(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.BulkImportService/Checkpoint',
            gqldb__pb2.CheckpointRequest.SerializeToString,
            gqldb__pb2.CheckpointResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EndBulkImport(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.BulkImportService/EndBulkImport',
            gqldb__pb2.EndBulkImportRequest.SerializeToString,
            gqldb__pb2.EndBulkImportResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def AbortBulkImport(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.BulkImportService/AbortBulkImport',
            gqldb__pb2.AbortBulkImportRequest.SerializeToString,
            gqldb__pb2.AbortBulkImportResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetBulkImportStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/gqldb.BulkImportService/GetBulkImportStatus',
            gqldb__pb2.GetBulkImportStatusRequest.SerializeToString,
            gqldb__pb2.GetBulkImportStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""Session service handles authentication and session lifecycle."""

from __future__ import annotations

from typing import TYPE_CHECKING

import grpc

if TYPE_CHECKING:
    from ..session import Session
    from .service_context import ServiceContext


class SessionService:
    """Session service for authentication and session management."""

    def __init__(self, ctx: ServiceContext):
        """Initialize session service.

        Args:
            ctx: Service context with shared dependencies
        """
        self.ctx = ctx
        self._stub = None

    def _get_stub(self):
        """Get or initialize session stub."""
        if self._stub is None:
            from ..proto import gqldb_pb2_grpc

            channel = self.ctx.get_channel()
            self._stub = gqldb_pb2_grpc.SessionServiceStub(channel)
        return self._stub

    def login(self, username: str, password: str, default_graph: str) -> Session:
        """Authenticate the user and create a session.

        Args:
            username: Username for authentication
            password: Password for authentication
            default_graph: Default graph name

        Returns:
            The created Session

        Raises:
            GqldbError: If login fails
        """
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        try:
            request = gqldb_pb2.LoginRequest(
                username=username,
                password=password,
                default_graph=default_graph
            )
            response = self._get_stub().Login(
                request,
                timeout=self.ctx.config.timeout
            )
            return self.ctx.sessions.login(
                response.session_id,
                response.server_version,
                list(response.roles),
                default_graph
            )
        except grpc.RpcError as e:
            raise GqldbError(f"Login failed: {e.details()}") from e

    def logout(self) -> None:
        """Terminate the current session.

        Raises:
            GqldbError: If logout fails
        """
        if not self.ctx.sessions.is_logged_in():
            return

        if self._stub:
            from ..proto import gqldb_pb2

            try:
                metadata = self.ctx.get_session_metadata()
                request = gqldb_pb2.LogoutRequest()
                self._stub.Logout(
                    request,
                    metadata=metadata,
                    timeout=self.ctx.config.timeout
                )
            except grpc.RpcError:
                pass  # Best effort logout

        self.ctx.tx_manager.clear_all()
        self.ctx.sessions.logout()

    def ping(self) -> int:
        """Check the connection and return the latency in nanoseconds.

        Returns:
            Latency in nanoseconds

        Raises:
            GqldbError: If ping fails
        """
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        try:
            metadata = self.ctx.get_session_metadata()
            request = gqldb_pb2.PingRequest()
            response = self._get_stub().Ping(
                request,
                metadata=metadata,
                timeout=self.ctx.config.timeout
            )
            self.ctx.update_activity()
            return response.latency_ns
        except grpc.RpcError as e:
            raise GqldbError(f"Ping failed: {e.details()}") from e



"""Services module for GQLDB Python driver."""

from .service_context import ServiceContext
from .converters import (
    convert_gql_response,
    convert_graph_info,
    convert_health_status,
    node_data_to_proto,
    edge_data_to_proto,
    convert_proto_properties,
)
from .session_service import SessionService
from .query_service import QueryService
from .graph_service import GraphService
from .transaction_service import TransactionService
from .data_service import DataService
from .health_service import HealthService
from .admin_service import AdminService
from .bulk_import_service import BulkImportService

__all__ = [
    'ServiceContext',
    'convert_gql_response',
    'convert_graph_info',
    'convert_health_status',
    'node_data_to_proto',
    'edge_data_to_proto',
    'convert_proto_properties',
    'SessionService',
    'QueryService',
    'GraphService',
    'TransactionService',
    'DataService',
    'HealthService',
    'AdminService',
    'BulkImportService',
]



"""Pytest configuration and fixtures for GQLDB tests."""

import pytest

from gqldb.config import ConfigBuilder, GqldbConfig
from gqldb.client import GqldbClient
from gqldb.errors import GqldbError


# Test database configuration
AUTH_HOST = "192.168.1.104:60061"      # Server with authentication
NO_AUTH_HOST = "192.168.1.104:60062"   # Server without authentication
TEST_HOST = AUTH_HOST                  # Default to auth server
TEST_USERNAME = "admin"
TEST_PASSWORD = "root11"
TEST_DATABASE = "miniCircle"

# Flag to track if authentication is enabled on the server
AUTH_ENABLED = None


def check_auth_enabled(client: GqldbClient) -> bool:
    """Check if authentication is enabled on the server."""
    global AUTH_ENABLED
    if AUTH_ENABLED is not None:
        return AUTH_ENABLED

    try:
        client.login(TEST_USERNAME, TEST_PASSWORD)
        client.logout()
        AUTH_ENABLED = True
    except GqldbError as e:
        if "authentication is not enabled" in str(e).lower():
            AUTH_ENABLED = False
        else:
            # Some other error, assume auth is enabled
            AUTH_ENABLED = True
    return AUTH_ENABLED


def drop_test_graph(client: GqldbClient, graph_name: str) -> None:
    """Drop a test graph safely. Switches to miniCircle first."""
    try:
        client.use_graph(TEST_DATABASE)
    except Exception:
        pass
    try:
        client.drop_graph(graph_name, if_exists=True)
    except Exception:
        pass


def cleanup_test_graphs(client: GqldbClient) -> None:
    """Clean up leftover test graphs to avoid hitting 3-graph limit."""
    try:
        graphs = client.list_graphs()
        for g in graphs:
            if g.name.startswith("test_"):
                print(f"Cleaning up leftover test graph: {g.name}")
                drop_test_graph(client, g.name)
    except Exception:
        pass


@pytest.fixture
def test_config() -> GqldbConfig:
    """Return test configuration."""
    return (
        ConfigBuilder()
        .hosts(TEST_HOST)
        .username(TEST_USERNAME)
        .password(TEST_PASSWORD)
        .default_graph(TEST_DATABASE)
        .timeout(30)
        .build()
    )


@pytest.fixture
def client(test_config: GqldbConfig) -> GqldbClient:
    """Return a GQLDB client (not logged in)."""
    return GqldbClient(test_config)


@pytest.fixture
def connected_client(test_config: GqldbConfig) -> GqldbClient:
    """Return a GQLDB client that is ready to use.

    If authentication is enabled, logs in first.
    If authentication is disabled, just returns the client.
    """
    client = GqldbClient(test_config)

    if check_auth_enabled(client):
        client.login(TEST_USERNAME, TEST_PASSWORD)

    cleanup_test_graphs(client)

    yield client

    try:
        if client.is_logged_in():
            client.logout()
    except Exception:
        pass
    client.close()


@pytest.fixture
def logged_in_client(client: GqldbClient) -> GqldbClient:
    """Return a logged-in GQLDB client (skips if auth disabled)."""
    if not check_auth_enabled(client):
        pytest.skip("Authentication is not enabled on this server")

    client.login(TEST_USERNAME, TEST_PASSWORD)
    yield client
    try:
        client.logout()
    except Exception:
        pass
    client.close()


@pytest.fixture
def no_auth_config() -> GqldbConfig:
    """Return configuration for no-auth server."""
    return (
        ConfigBuilder()
        .hosts(NO_AUTH_HOST)
        .default_graph(TEST_DATABASE)
        .timeout(30)
        .build()
    )


@pytest.fixture
def no_auth_client(no_auth_config: GqldbConfig) -> GqldbClient:
    """Return a GQLDB client connected to no-auth server."""
    client = GqldbClient(no_auth_config)
    yield client
    client.close()



"""Query service handles GQL query execution."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Callable, Optional

import grpc

if TYPE_CHECKING:
    from ..client import QueryConfig
    from ..response import Response
    from .service_context import ServiceContext


class QueryService:
    """Query service for executing GQL queries."""

    def __init__(self, ctx: ServiceContext):
        """Initialize query service.

        Args:
            ctx: Service context with shared dependencies
        """
        self.ctx = ctx
        self._stub = None

    def _calculate_timeout(self, config: Optional[QueryConfig] = None) -> int:
        """Calculate timeout with the following priority:
        1. QueryConfig.timeout (highest priority) - explicitly specified timeout
        2. gRPC context deadline (medium priority) - remaining time from context deadline
        3. client.config.timeout (default) - default timeout from client config

        Args:
            config: Optional query configuration

        Returns:
            Timeout in seconds
        """
        from ..client import QueryConfig as QC
        cfg = config or QC()

        # Priority 1: Use explicitly specified timeout from QueryConfig
        if cfg.timeout and cfg.timeout > 0:
            return cfg.timeout

        # Priority 2: Try to get deadline from gRPC context (if available)
        # Note: This requires the caller to pass a context with deadline
        # For now, we skip this as Python SDK doesn't expose context parameter yet

        # Priority 3: Use default timeout from client config
        return self.ctx.config.timeout

    def _get_stub(self):
        """Get or initialize query stub."""
        # WORKAROUND: Bypass ConnectionPool for long-running queries
        # ConnectionPool has a 30-second timeout issue
        import grpc
        from ..proto import gqldb_pb2_grpc

        # Get host from config
        host = self.ctx.config.hosts[0] if self.ctx.config.hosts else "localhost:60061"

        # Create channel directly (bypassing ConnectionPool)
        # TODO: Fix ConnectionPool 30-second timeout issue
        channel = grpc.insecure_channel(host)
        return gqldb_pb2_grpc.QueryServiceStub(channel)

    def gql(self, query: str, config: Optional[QueryConfig] = None) -> Response:
        """Execute a GQL query and return the result.

        Args:
            query: GQL query string
            config: Optional query configuration

        Returns:
            Query response

        Raises:
            EmptyQueryError: If query is empty
            GqldbError: If query fails
        """
        from ..client import QueryConfig as QC
        from ..errors import EmptyQueryError, GqldbError
        from ..proto import gqldb_pb2
        from .converters import convert_gql_response

        if not query:
            raise EmptyQueryError()

        session = self.ctx.sessions.get_session()
        cfg = config or QC()
        timeout = self._calculate_timeout(config)

        try:
            # Handle graph name with priority: config > session > client config
            graph_name = cfg.graph_name
            if not graph_name:
                graph_name = self.ctx.sessions.get_default_graph()
            if not graph_name:
                graph_name = self.ctx.config.default_graph

            request = gqldb_pb2.GqlRequest(
                gql=query,
                graph_name=graph_name or '',
                session_id=session.id if session else 0,
                transaction_id=cfg.transaction_id,
                timeout=timeout,
                read_only=cfg.read_only,
                max_path_results=cfg.max_path_results
            )

            metadata = self.ctx.get_session_metadata()
            proto_response = self._get_stub().Gql(
                request,
                metadata=metadata,
                timeout=timeout
            )

            self.ctx.update_activity()
            return convert_gql_response(proto_response)
        except grpc.RpcError as e:
            raise GqldbError(f"Query failed: {e.details()}") from e

    def gql_stream(
        self,
        query: str,
        config: Optional[QueryConfig] = None,
        callback: Optional[Callable[[Response], None]] = None
    ) -> None:
        """Execute a GQL query and stream the results.

        Args:
            query: GQL query string
            config: Optional query configuration
            callback: Callback function for each response chunk

        Raises:
            EmptyQueryError: If query is empty
            GqldbError: If query fails
        """
        from ..client import QueryConfig as QC
        from ..errors import EmptyQueryError, GqldbError
        from ..proto import gqldb_pb2
        from .converters import convert_gql_response

        if not query:
            raise EmptyQueryError()

        session = self.ctx.sessions.get_session()
        cfg = config or QC()
        timeout = self._calculate_timeout(config)

        try:
            # Handle graph name with priority: config > session > client config
            graph_name = cfg.graph_name
            if not graph_name:
                graph_name = self.ctx.sessions.get_default_graph()
            if not graph_name:
                graph_name = self.ctx.config.default_graph

            request = gqldb_pb2.GqlRequest(
                gql=query,
                graph_name=graph_name or '',
                session_id=session.id if session else 0,
                transaction_id=cfg.transaction_id,
                timeout=timeout,
                read_only=cfg.read_only,
                max_path_results=cfg.max_path_results
            )

            metadata = self.ctx.get_session_metadata()
            stream = self._get_stub().GqlStream(
                request,
                metadata=metadata,
                timeout=timeout
            )

            for proto_response in stream:
                self.ctx.update_activity()
                response = convert_gql_response(proto_response)
                if callback:
                    callback(response)
        except grpc.RpcError as e:
            raise GqldbError(f"Stream query failed: {e.details()}") from e

    def explain(self, query: str, config: Optional[QueryConfig] = None) -> str:
        """Return the execution plan for a query.

        Args:
            query: GQL query string
            config: Optional query configuration

        Returns:
            Execution plan string

        Raises:
            EmptyQueryError: If query is empty
            GqldbError: If explain fails
        """
        from ..client import QueryConfig as QC
        from ..errors import EmptyQueryError, GqldbError
        from ..proto import gqldb_pb2

        if not query:
            raise EmptyQueryError()

        session = self.ctx.sessions.get_session()
        cfg = config or QC()
        timeout = self._calculate_timeout(config)

        try:
            # Handle graph name with priority: config > session > client config
            graph_name = cfg.graph_name
            if not graph_name:
                graph_name = self.ctx.sessions.get_default_graph()
            if not graph_name:
                graph_name = self.ctx.config.default_graph

            request = gqldb_pb2.GqlRequest(
                gql=query,
                graph_name=graph_name or '',
                session_id=session.id if session else 0,
                transaction_id=cfg.transaction_id,
                timeout=timeout,
                read_only=cfg.read_only,
                max_path_results=cfg.max_path_results
            )

            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().Explain(
                request,
                metadata=metadata,
                timeout=timeout
            )

            return response.plan
        except grpc.RpcError as e:
            raise GqldbError(f"Explain failed: {e.details()}") from e

    def profile(self, query: str, config: Optional[QueryConfig] = None) -> str:
        """Execute a query with profiling and return statistics.

        Args:
            query: GQL query string
            config: Optional query configuration

        Returns:
            Profiling statistics string

        Raises:
            EmptyQueryError: If query is empty
            GqldbError: If profile fails
        """
        from ..client import QueryConfig as QC
        from ..errors import EmptyQueryError, GqldbError
        from ..proto import gqldb_pb2

        if not query:
            raise EmptyQueryError()

        session = self.ctx.sessions.get_session()
        cfg = config or QC()
        timeout = self._calculate_timeout(config)

        try:
            # Handle graph name with priority: config > session > client config
            graph_name = cfg.graph_name
            if not graph_name:
                graph_name = self.ctx.sessions.get_default_graph()
            if not graph_name:
                graph_name = self.ctx.config.default_graph

            request = gqldb_pb2.GqlRequest(
                gql=query,
                graph_name=graph_name or '',
                session_id=session.id if session else 0,
                transaction_id=cfg.transaction_id,
                timeout=timeout,
                read_only=cfg.read_only,
                max_path_results=cfg.max_path_results
            )

            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().Profile(
                request,
                metadata=metadata,
                timeout=timeout
            )

            return response.profile
        except grpc.RpcError as e:
            raise GqldbError(f"Profile failed: {e.details()}") from e



"""Pretty printers for GQLDB data types."""

from typing import Any, Dict, List, Optional

from tabulate import tabulate

from .response import Response, Node, Edge, Path
from .types import Schema, Table


def print_nodes(nodes: List[Node], schemas: Optional[Dict[str, Schema]] = None) -> str:
    """Print nodes in tabular format grouped by schema.

    Args:
        nodes: List of nodes to print.
        schemas: Optional schema map for property ordering.

    Returns:
        Formatted string representation.
    """
    if not nodes:
        return "No nodes found."

    result = []
    last_schema = ""
    current_rows = []
    current_headers = []

    for node in nodes:
        if node is None:
            continue

        # Get the first label as schema name
        schema_name = node.labels[0] if node.labels else ""

        # Switch schema if needed
        if schema_name != last_schema:
            # Print previous table if exists
            if current_rows:
                result.append(tabulate(current_rows, headers=current_headers, tablefmt="grid"))
                result.append("")

            # Build new header
            current_headers = ["ID", "Labels"]
            if schemas and schema_name in schemas:
                for prop in schemas[schema_name].properties:
                    current_headers.append(prop.name)
            else:
                # Derive from node properties
                for key in node.properties.keys():
                    current_headers.append(key)

            current_rows = []
            last_schema = schema_name

        # Build row
        row = [node.id, _format_labels(node.labels)]
        if schemas and schema_name in schemas:
            for prop in schemas[schema_name].properties:
                row.append(_format_value(node.properties.get(prop.name)))
        else:
            for key in node.properties.keys():
                row.append(_format_value(node.properties.get(key)))

        current_rows.append(row)

    # Print last table
    if current_rows:
        result.append(tabulate(current_rows, headers=current_headers, tablefmt="grid"))

    result.append(f"\n{len(nodes)} node(s)")
    return "\n".join(result)


def print_nodes_without_schema(nodes: List[Node]) -> str:
    """Print nodes deriving schema from data.

    Args:
        nodes: List of nodes to print.

    Returns:
        Formatted string representation.
    """
    if not nodes:
        return "No nodes found."

    # Build schemas from nodes
    schemas: Dict[str, Schema] = {}
    for node in nodes:
        if node is None:
            continue
        for label in node.labels:
            if label not in schemas:
                schemas[label] = _build_schema_from_node(label, node)

    return print_nodes(nodes, schemas)


def print_edges(edges: List[Edge], schemas: Optional[Dict[str, Schema]] = None) -> str:
    """Print edges in tabular format grouped by schema.

    Args:
        edges: List of edges to print.
        schemas: Optional schema map for property ordering.

    Returns:
        Formatted string representation.
    """
    if not edges:
        return "No edges found."

    result = []
    last_schema = ""
    current_rows = []
    current_headers = []

    for edge in edges:
        if edge is None:
            continue

        schema_name = edge.label

        # Switch schema if needed
        if schema_name != last_schema:
            # Print previous table if exists
            if current_rows:
                result.append(tabulate(current_rows, headers=current_headers, tablefmt="grid"))
                result.append("")

            # Build new header
            current_headers = ["ID", "FROM", "TO", "Label"]
            if schemas and schema_name in schemas:
                for prop in schemas[schema_name].properties:
                    current_headers.append(prop.name)
            else:
                # Derive from edge properties
                for key in edge.properties.keys():
                    current_headers.append(key)

            current_rows = []
            last_schema = schema_name

        # Build row
        row = [edge.id, edge.from_node_id, edge.to_node_id, edge.label]
        if schemas and schema_name in schemas:
            for prop in schemas[schema_name].properties:
                row.append(_format_value(edge.properties.get(prop.name)))
        else:
            for key in edge.properties.keys():
                row.append(_format_value(edge.properties.get(key)))

        current_rows.append(row)

    # Print last table
    if current_rows:
        result.append(tabulate(current_rows, headers=current_headers, tablefmt="grid"))

    result.append(f"\n{len(edges)} edge(s)")
    return "\n".join(result)


def print_edges_without_schema(edges: List[Edge]) -> str:
    """Print edges deriving schema from data.

    Args:
        edges: List of edges to print.

    Returns:
        Formatted string representation.
    """
    if not edges:
        return "No edges found."

    # Build schemas from edges
    schemas: Dict[str, Schema] = {}
    for edge in edges:
        if edge is None:
            continue
        if edge.label and edge.label not in schemas:
            schemas[edge.label] = _build_schema_from_edge(edge.label, edge)

    return print_edges(edges, schemas)


def print_paths(paths: List[Path]) -> str:
    """Print paths with node-edge sequence notation.

    Args:
        paths: List of paths to print.

    Returns:
        Formatted string representation.
    """
    if not paths:
        return "No paths found."

    rows = []
    for i, path in enumerate(paths):
        if path is None:
            continue
        path_str = _format_path(path)
        rows.append([i, path_str])

    result = tabulate(rows, headers=["#", "Path"], tablefmt="grid")
    return result + f"\n{len(paths)} path(s)"


def print_table(table: Table) -> str:
    """Print generic table data.

    Args:
        table: Table to print.

    Returns:
        Formatted string representation.
    """
    if table is None or not table.headers:
        return "No table data."

    headers = [h.name for h in table.headers]
    rows = [[_format_value(v) for v in row] for row in table.rows]

    result = []
    if table.name:
        result.append(f"Table: {table.name}")

    result.append(tabulate(rows, headers=headers, tablefmt="grid"))
    result.append(f"\n{len(table.rows)} row(s)")
    return "\n".join(result)


def print_any(resp: Response) -> str:
    """Auto-detect type from Response and print accordingly.

    Args:
        resp: Response to print.

    Returns:
        Formatted string representation.
    """
    if resp is None or resp.is_empty():
        return "No data."

    # Try to auto-detect the type from the first column
    if resp.columns:
        try:
            ar = resp.get(0)

            # Try edges first
            try:
                edges, schemas = ar.as_edges()
                if edges:
                    return print_edges(edges, schemas)
            except (TypeError, Exception):
                pass

            # Try nodes
            try:
                nodes, schemas = ar.as_nodes()
                if nodes:
                    return print_nodes(nodes, schemas)
            except (TypeError, Exception):
                pass

            # Try paths
            try:
                paths = ar.as_paths()
                if paths:
                    return print_paths(paths)
            except (TypeError, Exception):
                pass
        except Exception:
            pass

    # Default to table format
    ar = resp.get(0)
    table = ar.as_table()
    return print_table(table)


# Helper functions

def _format_labels(labels: List[str]) -> str:
    """Format labels for display."""
    if not labels:
        return ""
    return "[" + ", ".join(labels) + "]"


def _format_value(v: Any) -> str:
    """Format a value for display."""
    if v is None:
        return "<nil>"
    if isinstance(v, list):
        return "[" + ", ".join(_format_value(item) for item in v) + "]"
    if isinstance(v, dict):
        parts = [f"{k}: {_format_value(val)}" for k, val in v.items()]
        return "{" + ", ".join(parts) + "}"
    return str(v)


def _format_path(path: Path) -> str:
    """Format a path for display."""
    if path is None:
        return "<nil>"

    parts = []
    node_idx = 0
    edge_idx = 0

    # Alternate between nodes and edges
    while node_idx < len(path.nodes) or edge_idx < len(path.edges):
        if node_idx < len(path.nodes):
            node = path.nodes[node_idx]
            label = ""
            if node.labels:
                label = ":" + node.labels[0]
            parts.append(f"({node.id}{label})")
            node_idx += 1

        if edge_idx < len(path.edges):
            edge = path.edges[edge_idx]
            parts.append(f"-[{edge.id}:{edge.label}]->")
            edge_idx += 1

    return " ".join(parts)


def _contains_any(lst: List[str], *items: str) -> bool:
    """Check if list contains any of the items."""
    for s in lst:
        if s in items:
            return True
    return False


def _build_schema_from_node(label: str, node: Node) -> Schema:
    """Build schema from a node."""
    from .types import PropertyDef, PropertyType
    properties = []
    for name, value in node.properties.items():
        properties.append(PropertyDef(name=name, type=_infer_property_type(value)))
    return Schema(name=label, properties=properties)


def _build_schema_from_edge(label: str, edge: Edge) -> Schema:
    """Build schema from an edge."""
    from .types import PropertyDef, PropertyType
    properties = []
    for name, value in edge.properties.items():
        properties.append(PropertyDef(name=name, type=_infer_property_type(value)))
    return Schema(name=label, properties=properties)


def _infer_property_type(value: Any):
    """Infer property type from value."""
    from .types import PropertyType
    if value is None:
        return PropertyType.NULL
    if isinstance(value, bool):
        return PropertyType.BOOL
    if isinstance(value, int):
        return PropertyType.INT64
    if isinstance(value, float):
        return PropertyType.DOUBLE
    if isinstance(value, str):
        return PropertyType.STRING
    if isinstance(value, bytes):
        return PropertyType.BLOB
    if isinstance(value, list):
        return PropertyType.LIST
    if isinstance(value, dict):
        return PropertyType.MAP
    return PropertyType.UNSET



"""End-to-end integration tests for GQLDB Python SDK.

Complete workflow tests for real-world scenarios.
Run with: pytest tests/e2e/ -v
"""

import os
import time
import threading
import pytest

from gqldb.config import ConfigBuilder, GqldbConfig
from gqldb.client import GqldbClient
from gqldb.types import NodeData
from gqldb.errors import GqldbError


# Test configuration
HOST = os.getenv("GQLDB_HOST", "192.168.1.104:60061")
USERNAME = os.getenv("GQLDB_USERNAME", "admin")
PASSWORD = os.getenv("GQLDB_PASSWORD", "root11")
GRAPH = os.getenv("GQLDB_TEST_GRAPH", "miniCircle")


def create_client() -> GqldbClient:
    """Create a new client instance."""
    config = (
        ConfigBuilder()
        .hosts(HOST)
        .username(USERNAME)
        .password(PASSWORD)
        .default_graph(GRAPH)
        .timeout(60)
        .build()
    )
    return GqldbClient(config)


def unique_graph_name(prefix: str) -> str:
    """Generate a unique graph name."""
    return f"{prefix}_{int(time.time() * 1000)}"


# ==================== Complete CRUD Lifecycle ====================


@pytest.mark.e2e
class TestCompleteCRUDLifecycle:
    """Tests for complete CRUD workflow."""

    def test_crud_lifecycle(self):
        """Test the full create-read-update-delete workflow."""
        client = create_client()

        try:
            client.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        graph_name = unique_graph_name("e2e_crud")

        try:
            # 1. Create graph
            client.gql(f"CREATE GRAPH {graph_name}")

            # 2. Create schema
            try:
                client.gql(f"USE GRAPH {graph_name} CREATE NODE TYPE Person (name STRING, age INT64)")
            except GqldbError:
                pass  # Schema may already exist

            # 3. Insert nodes
            session = client.start_bulk_import(graph_name)
            nodes = [
                NodeData(labels=["Person"], properties={"name": "Alice", "age": 30}),
                NodeData(labels=["Person"], properties={"name": "Bob", "age": 25}),
                NodeData(labels=["Person"], properties={"name": "Charlie", "age": 35}),
            ]
            try:
                client.insert_nodes(graph_name, nodes,
                                   bulk_import_session_id=session.session_id)
            finally:
                client.end_bulk_import(session.session_id)

            # 4. Query nodes
            response = client.gql(f"USE GRAPH {graph_name} MATCH (p:Person) RETURN p.name, p.age ORDER BY p.name")
            assert not response.is_empty()

            count = 0
            for row in response:
                count += 1
            assert count == 3

            # 5. Create edges
            try:
                client.gql(f"USE GRAPH {graph_name} CREATE EDGE TYPE KNOWS")
            except GqldbError:
                pass

            client.gql(f"""
                USE GRAPH {graph_name}
                INSERT (:Person {{name: 'Alice'}})-[:KNOWS]->(:Person {{name: 'Bob'}})
            """)

            # 6. Query paths
            path_response = client.gql(f"USE GRAPH {graph_name} MATCH p = (a:Person)-[e:KNOWS]->(b:Person) RETURN p LIMIT 10")
            paths = path_response.alias("p").as_paths()

            # 7. Delete node
            client.gql(f"USE GRAPH {graph_name} MATCH (p:Person WHERE p.name = 'Charlie') DELETE p")

            # 8. Verify
            verify_response = client.gql(f"USE GRAPH {graph_name} MATCH (p:Person) RETURN count(p) AS cnt")
            assert not verify_response.is_empty()

        finally:
            # Cleanup
            try:
                client.gql(f"DROP GRAPH {graph_name}")
            except GqldbError:
                pass
            client.close()


# ==================== Transaction Isolation ====================


@pytest.mark.e2e
class TestTransactionIsolation:
    """Tests for transaction isolation."""

    def test_transaction_isolation(self):
        """Test that transactions are properly isolated."""
        client1 = create_client()
        client2 = create_client()

        try:
            client1.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        try:
            client2.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        graph_name = unique_graph_name("e2e_txiso")

        try:
            # Create test graph
            client1.gql(f"CREATE GRAPH {graph_name}")

            # Begin transaction on client 1
            tx1 = client1.begin_transaction(graph_name, read_only=False)

            # Insert data in transaction
            try:
                from gqldb.client import QueryConfig
                client1.gql(f"USE GRAPH {graph_name} INSERT (:IsolationTest {{value: 'tx1'}})",
                           config=QueryConfig(transaction_id=tx1.id))
            except GqldbError:
                pass

            # Client 2 queries (may or may not see uncommitted data)
            try:
                response2 = client2.gql(f"USE GRAPH {graph_name} MATCH (n:IsolationTest) RETURN n")
            except GqldbError:
                pass

            # Commit
            client1.commit(tx1.id)

            # Client 2 should see data now
            try:
                response2_after = client2.gql(f"USE GRAPH {graph_name} MATCH (n:IsolationTest) RETURN n")
            except GqldbError:
                pass

        finally:
            try:
                client1.gql(f"DROP GRAPH {graph_name}")
            except GqldbError:
                pass
            client1.close()
            client2.close()


# ==================== Bulk Import Workflow ====================


@pytest.mark.e2e
class TestBulkImportWorkflow:
    """Tests for bulk import workflow."""

    def test_bulk_import_workflow(self):
        """Test the complete bulk import process."""
        client = create_client()

        try:
            client.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        graph_name = unique_graph_name("e2e_bulk")

        try:
            # Create graph
            client.gql(f"CREATE GRAPH {graph_name}")

            # Start bulk import
            session = client.start_bulk_import(graph_name)
            if session is None:
                pytest.skip("Bulk import not supported")
                return

            # Insert in batches
            batch_size = 100
            total_nodes = 500

            for i in range(0, total_nodes, batch_size):
                nodes = [
                    NodeData(labels=["BulkNode"], properties={"id": i + j, "batch": i // batch_size})
                    for j in range(batch_size)
                ]

                try:
                    client.insert_nodes(graph_name, nodes,
                                       bulk_import_session_id=session.session_id)
                except GqldbError:
                    pass

            # End session
            try:
                client.end_bulk_import(session.session_id)
            except GqldbError:
                pass

            # Verify
            try:
                response = client.gql(f"USE GRAPH {graph_name} MATCH (n:BulkNode) RETURN count(n) AS cnt")
            except GqldbError:
                pass

        finally:
            try:
                client.gql(f"DROP GRAPH {graph_name}")
            except GqldbError:
                pass
            client.close()


# ==================== Concurrent Operations ====================


@pytest.mark.e2e
class TestConcurrentOperations:
    """Tests for concurrent operations."""

    def test_concurrent_queries(self):
        """Test multiple clients performing parallel queries."""
        num_clients = 5
        queries_per_client = 10
        errors = []
        lock = threading.Lock()

        def client_worker(client_id: int):
            client = create_client()
            try:
                client.login(USERNAME, PASSWORD)
            except GqldbError as e:
                if "authentication is not enabled" not in str(e).lower():
                    with lock:
                        errors.append(f"Client {client_id} login: {e}")
                    return

            try:
                for j in range(queries_per_client):
                    try:
                        client.gql(f"RETURN {client_id} AS client, {j} AS query")
                    except GqldbError as e:
                        with lock:
                            errors.append(f"Client {client_id} query {j}: {e}")
            finally:
                client.close()

        threads = []
        for i in range(num_clients):
            t = threading.Thread(target=client_worker, args=(i,))
            threads.append(t)
            t.start()

        for t in threads:
            t.join(timeout=60)

        # Report errors but don't fail if some queries failed
        for error in errors:
            print(f"Concurrent error: {error}")


# ==================== Reconnection Scenario ====================


@pytest.mark.e2e
class TestReconnection:
    """Tests for reconnection scenarios."""

    def test_reconnection(self):
        """Test client reconnection after connection loss."""
        client = create_client()

        try:
            client.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        # Execute query
        client.gql("RETURN 1")

        # Close connection
        client.close()

        # Recreate and reconnect
        client = create_client()
        try:
            client.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        # Execute again
        response = client.gql("RETURN 2")
        assert not response.is_empty()

        client.close()



"""Admin service handles administrative operations."""

from __future__ import annotations

from typing import TYPE_CHECKING

import grpc

if TYPE_CHECKING:
    from ..types import CacheStats, CacheType, Statistics
    from .service_context import ServiceContext


class AdminService:
    """Admin service for administrative operations."""

    def __init__(self, ctx: ServiceContext):
        self.ctx = ctx
        self._stub = None

    def _get_stub(self):
        if self._stub is None:
            from ..proto import gqldb_pb2_grpc
            channel = self.ctx.get_channel()
            self._stub = gqldb_pb2_grpc.AdminServiceStub(channel)
        return self._stub

    def warmup_parser(self, count: int) -> None:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        try:
            request = gqldb_pb2.WarmupParserRequest(count=count)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().WarmupParser(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            if not response.success:
                raise GqldbError(response.message)
        except grpc.RpcError as e:
            raise GqldbError(f"Warmup parser failed: {e.details()}") from e

    def get_cache_stats(self, cache_type: CacheType) -> CacheStats:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import ASTCacheStats, CacheStats, PlanCacheStats

        try:
            request = gqldb_pb2.GetCacheStatsRequest(cache_type=cache_type.value)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().GetCacheStats(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )

            ast_stats = None
            plan_stats = None

            if response.ast_stats:
                ast_stats = ASTCacheStats(
                    hits=response.ast_stats.hits,
                    misses=response.ast_stats.misses,
                    evictions=response.ast_stats.evictions,
                    entries=response.ast_stats.entries,
                    hit_rate=response.ast_stats.hit_rate
                )

            if response.plan_stats:
                plan_stats = PlanCacheStats(
                    size=response.plan_stats.size,
                    capacity=response.plan_stats.capacity,
                    hits=response.plan_stats.hits,
                    misses=response.plan_stats.misses,
                    hit_rate=response.plan_stats.hit_rate
                )

            return CacheStats(ast_stats=ast_stats, plan_stats=plan_stats)
        except grpc.RpcError as e:
            raise GqldbError(f"Get cache stats failed: {e.details()}") from e

    def clear_cache(self, cache_type: CacheType) -> None:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        try:
            request = gqldb_pb2.ClearCacheRequest(cache_type=cache_type.value)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().ClearCache(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            if not response.success:
                raise GqldbError(response.message)
        except grpc.RpcError as e:
            raise GqldbError(f"Clear cache failed: {e.details()}") from e

    def get_statistics(self, graph_name: str) -> Statistics:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import Statistics

        try:
            request = gqldb_pb2.GetStatisticsRequest(graph_name=graph_name)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().GetStatistics(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )

            return Statistics(
                node_count=response.node_count,
                edge_count=response.edge_count,
                label_counts=dict(response.label_counts),
                edge_label_counts=dict(response.edge_label_counts)
            )
        except grpc.RpcError as e:
            raise GqldbError(f"Get statistics failed: {e.details()}") from e

    def invalidate_permission_cache(self, username: str) -> None:
        from ..errors import GqldbError
        from ..proto import gqldb_pb2

        try:
            request = gqldb_pb2.InvalidatePermissionCacheRequest(username=username)
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().InvalidatePermissionCache(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            if not response.success:
                raise GqldbError(response.message)
        except grpc.RpcError as e:
            raise GqldbError(f"Invalidate permission cache failed: {e.details()}") from e

    def compact(self) -> "CompactResult":
        """Trigger manual compaction of the database storage.

        Returns:
            CompactResult with success status and message.

        Raises:
            GqldbError: If compact fails.
        """
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import CompactResult

        try:
            request = gqldb_pb2.CompactRequest()
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().Compact(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            return CompactResult(
                success=response.success,
                message=response.message
            )
        except grpc.RpcError as e:
            raise GqldbError(f"Compact failed: {e.details()}") from e

    def wait_for_compute_topology(self, graph_name: str, timeout_ms: int = 0) -> "ComputeTopologyResult":
        """Wait for the computing engine topology to be ready.

        Args:
            graph_name: Graph to check.
            timeout_ms: Timeout in milliseconds (0 = check current status only).

        Returns:
            ComputeTopologyResult with ready status and message.

        Raises:
            GqldbError: If the call fails.
        """
        from ..errors import GqldbError
        from ..proto import gqldb_pb2
        from ..types import ComputeTopologyResult

        try:
            request = gqldb_pb2.WaitForComputeTopologyRequest(
                graph_name=graph_name,
                timeout_ms=timeout_ms
            )
            metadata = self.ctx.get_session_metadata()
            response = self._get_stub().WaitForComputeTopology(
                request, metadata=metadata, timeout=self.ctx.config.timeout
            )
            return ComputeTopologyResult(
                ready=response.ready,
                message=response.message
            )
        except grpc.RpcError as e:
            raise GqldbError(f"Wait for compute topology failed: {e.details()}") from e



"""Smoke tests for GQLDB Python SDK.

Quick sanity checks to verify basic SDK functionality.
Run with: pytest tests/smoke/ -v
"""

import os
import pytest

from gqldb.config import ConfigBuilder, GqldbConfig
from gqldb.client import GqldbClient
from gqldb.errors import GqldbError


# Test configuration - can be overridden via environment variables
HOST = os.getenv("GQLDB_HOST", "192.168.1.104:60061")
USERNAME = os.getenv("GQLDB_USERNAME", "admin")
PASSWORD = os.getenv("GQLDB_PASSWORD", "root11")
GRAPH = os.getenv("GQLDB_TEST_GRAPH", "miniCircle")


@pytest.fixture
def config() -> GqldbConfig:
    """Return test configuration."""
    return (
        ConfigBuilder()
        .hosts(HOST)
        .username(USERNAME)
        .password(PASSWORD)
        .default_graph(GRAPH)
        .timeout(30)
        .build()
    )


@pytest.fixture
def client(config: GqldbConfig) -> GqldbClient:
    """Return a GQLDB client."""
    c = GqldbClient(config)
    yield c
    c.close()


@pytest.mark.smoke
class TestSmokeTests:
    """Smoke tests for basic SDK functionality."""

    def test_client_creation(self, config: GqldbConfig):
        """Verify that a client can be instantiated successfully."""
        client = GqldbClient(config)
        assert client is not None
        client.close()

    def test_connection_ping(self, client: GqldbClient):
        """Verify that the client can ping the server."""
        # Try to login first (may fail if auth is disabled)
        try:
            client.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" in str(e).lower():
                pass  # Continue without login
            else:
                raise

        latency = client.ping()
        assert latency is not None
        assert latency > 0

    def test_authentication(self, client: GqldbClient):
        """Verify that login succeeds with valid credentials."""
        try:
            session_info = client.login(USERNAME, PASSWORD)
            assert session_info is not None
        except GqldbError as e:
            if "authentication is not enabled" in str(e).lower():
                pytest.skip("Authentication is not enabled on this server")
            raise

    def test_simple_query(self, client: GqldbClient):
        """Verify that a simple query can be executed."""
        # Login first
        try:
            client.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        response = client.gql("RETURN 1 AS result")
        assert response is not None
        assert not response.is_empty()

    def test_health_check(self, client: GqldbClient):
        """Verify that the health service responds."""
        # Login first
        try:
            client.login(USERNAME, PASSWORD)
        except GqldbError as e:
            if "authentication is not enabled" not in str(e).lower():
                raise

        status = client.health_check()
        assert status is not None

    def test_version_info(self):
        """Verify that the SDK version is accessible."""
        from gqldb import __version__
        assert __version__ is not None
        assert len(__version__) > 0



"""PropertyType integration tests.

Includes roundtrip verification for encode/decode correctness.
"""

import pytest
import uuid
from datetime import datetime, date
import math

from gqldb.client import GqldbClient, QueryConfig
from gqldb.types import GraphType, NodeData, EdgeData
from tests.conftest import drop_test_graph
from gqldb.types.data_types import (
    Point,
    Point3D,
    GqldbDecimal,
    GqldbLocalDateTime,
    GqldbZonedDateTime,
    GqldbLocalTime,
    GqldbZonedTime,
    YearToMonth,
    DayToSecond,
    Vector,
)


@pytest.mark.integration
class TestPropertyTypes:
    """Tests for all PropertyType values."""

    @pytest.fixture
    def propertytype_graph(self, connected_client: GqldbClient, request):
        """Create a temporary test graph for PropertyType tests."""
        graph_name = f"test_propertytype_{uuid.uuid4().hex[:8]}"
        connected_client.create_graph(graph_name, GraphType.OPEN, "PropertyType test graph")
        yield graph_name
        drop_test_graph(connected_client, graph_name)

    def test_insert_nodes_all_property_types(self, connected_client: GqldbClient, propertytype_graph: str):
        """Test inserting nodes with all supported PropertyTypes."""
        # Start bulk import session
        session = connected_client.start_bulk_import(propertytype_graph)

        try:
            now = datetime.now()
            today = date.today()

            nodes = [
                NodeData(
                    labels=["TestAllTypes"],
                    properties={
                        # Basic numeric types
                        "prop_int32": 42,
                        "prop_int64": 9223372036854775807,
                        "prop_float": 3.14,
                        "prop_double": 3.141592653589793,
                        # String types
                        "prop_string": "Hello, World!",
                        "prop_unicode": "你好世界 🌍",
                        # Boolean
                        "prop_bool_true": True,
                        "prop_bool_false": False,
                        # Time types
                        "prop_timestamp": now,
                        "prop_date": today,
                        # Binary
                        "prop_blob": bytes([0x00, 0x01, 0x02, 0xFF, 0xFE]),
                        # Complex types
                        "prop_list": ["a", "b", "c"],
                        "prop_list_mixed": [1, "two", 3.0],
                        "prop_map": {"key1": "value1", "key2": 2},
                        "prop_nested_list": [[1, 2], [3, 4]],
                    }
                ),
                NodeData(
                    labels=["TestNumericEdgeCases"],
                    properties={
                        # Edge cases for numeric types
                        "int32_min": -2147483648,
                        "int32_max": 2147483647,
                        "int64_min": -9223372036854775808,
                        "int64_max": 9223372036854775807,
                        "float_zero": 0.0,
                        "float_neg": -123.456,
                    }
                ),
                NodeData(
                    labels=["TestStringEdgeCases"],
                    properties={
                        "empty_string": "",
                        "special_chars": "!@#$%^&*()_+-=[]{}|;':\",./<>?",
                        "newlines": "line1\nline2\rline3",
                        "emoji_string": "😀🎉🚀💻🌟",
                        "japanese": "日本語テスト",
                        "arabic": "اختبار عربي",
                    }
                ),
            ]

            result = connected_client.insert_nodes(
                propertytype_graph,
                nodes,
                bulk_import_session_id=session.session_id
            )
            assert result.success
            assert result.node_count == 3

            print(f"Inserted {result.node_count} nodes with all PropertyTypes")
        finally:
            connected_client.end_bulk_import(session.session_id)

        # Verify by query
        config = QueryConfig(graph_name=propertytype_graph)
        response = connected_client.gql("MATCH (n:TestAllTypes) RETURN n", config)
        print(f"Verification query returned {len(response.rows)} rows")

    def test_insert_edges_all_property_types(self, connected_client: GqldbClient, propertytype_graph: str):
        """Test inserting edges with all supported PropertyTypes."""
        # Start bulk import session
        session = connected_client.start_bulk_import(propertytype_graph)

        try:
            # First insert nodes
            nodes = [
                NodeData(labels=["Source"], properties={"name": "SourceNode"}),
                NodeData(labels=["Target"], properties={"name": "TargetNode"}),
            ]
            node_result = connected_client.insert_nodes(
                propertytype_graph,
                nodes,
                bulk_import_session_id=session.session_id
            )
            assert node_result.success

            # Query to get node IDs
            config = QueryConfig(graph_name=propertytype_graph)
            resp = connected_client.gql(
                "MATCH (n) WHERE n.name IN ['SourceNode', 'TargetNode'] RETURN id(n) AS node_id, n.name ORDER BY n.name",
                config
            )

            node_ids = []
            for i in range(resp.row_count):
                row = resp.rows[i]
                value = resp.get_by_name(row, "node_id")
                if value:
                    node_ids.append(value)

            assert len(node_ids) >= 2
            from_id = node_ids[0]
            to_id = node_ids[1]
            now = datetime.now()

            edges = [
                EdgeData(
                    label="HAS_ALL_TYPES",
                    from_node_id=from_id,
                    to_node_id=to_id,
                    properties={
                        # Numeric types
                        "edge_int32": 100,
                        "edge_int64": 200,
                        "edge_double": 3.14159,
                        # String types
                        "edge_string": "edge property",
                        "edge_unicode": "边缘属性 🔗",
                        # Boolean
                        "edge_bool": True,
                        # Time
                        "edge_timestamp": now,
                        # Binary
                        "edge_blob": bytes([0xDE, 0xAD, 0xBE, 0xEF]),
                        # Complex types
                        "edge_list": ["x", "y", "z"],
                        "edge_map": {"weight": 1.5, "type": "strong"},
                    }
                ),
                EdgeData(
                    label="NUMERIC_EDGE",
                    from_node_id=from_id,
                    to_node_id=to_id,
                    properties={
                        "weight": 0.75,
                        "count": 42,
                        "distance": 1234.5678,
                    }
                ),
            ]

            result = connected_client.insert_edges(
                propertytype_graph,
                edges,
                bulk_import_session_id=session.session_id
            )
            assert result.success
            assert result.edge_count == 2

            print(f"Inserted {result.edge_count} edges with all PropertyTypes")
        finally:
            connected_client.end_bulk_import(session.session_id)


# ============================================================================
# Roundtrip Tests - Verify encode/decode correctness for all property types
# ============================================================================

@pytest.mark.integration
class TestPropertyTypeRoundtrip:
    """Roundtrip tests for all PropertyType values."""

    @pytest.fixture
    def roundtrip_graph(self, connected_client: GqldbClient):
        """Create a temporary test graph for roundtrip tests."""
        graph_name = f"test_roundtrip_{uuid.uuid4().hex[:8]}"
        connected_client.create_graph(graph_name, GraphType.OPEN, "Roundtrip test graph")
        yield graph_name
        drop_test_graph(connected_client, graph_name)

    def insert_and_query_property(
        self, client: GqldbClient, graph_name: str, prop_name: str, prop_value
    ):
        """Insert a node with a property and query it back."""
        # Start bulk import session
        session = client.start_bulk_import(graph_name)

        try:
            nodes = [NodeData(labels=["RoundtripTest"], properties={prop_name: prop_value})]
            result = client.insert_nodes(graph_name, nodes, bulk_import_session_id=session.session_id)
            assert result.success

            # Query to get node ID
            config = QueryConfig(graph_name=graph_name)
            query_for_id = f"MATCH (n:RoundtripTest) WHERE n.{prop_name} IS NOT NULL RETURN id(n) AS node_id ORDER BY id(n) DESC LIMIT 1"
            resp_id = client.gql(query_for_id, config)
            assert resp_id.row_count > 0

            node_id = resp_id.get_by_name(resp_id.rows[0], "node_id")

            # Query back the property
            query = f"MATCH (n:RoundtripTest) WHERE id(n) = '{node_id}' RETURN n.{prop_name} AS value"
            resp = client.gql(query, config)
            assert resp.row_count > 0

            return resp.get_by_name(resp.rows[0], "value")
        finally:
            client.end_bulk_import(session.session_id)

    # --- Numeric Types ---

    @pytest.mark.parametrize("idx,value", [
        (0, 0),
        (1, 9223372036854775807),
        (2, 1234567890123),
    ])
    def test_int64_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, value: int):
        """Test Int64 roundtrip."""
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"int64_{idx}", value)
        assert result == value

    @pytest.mark.parametrize("idx,value", [
        (0, 0.0),
        (1, 3.141592653589793),
        (2, 1e-300),
        (3, 1e300),
    ])
    def test_float64_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, value: float):
        """Test Float64 roundtrip."""
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"float64_{idx}", value)
        # Database may return int for integer-valued floats
        if isinstance(result, int):
            result = float(result)
        assert abs(result - value) < 1e-10

    # --- String Types ---

    @pytest.mark.parametrize("idx,value", [
        (0, ""),
        (1, "Hello, World!"),
        (2, "你好世界 🌍"),
        (3, "😀🎉🚀💻🌟"),
        (4, "!@#$%^&*()_+-=[]{}|;':\",./<>?"),
        (5, "line1\nline2\rline3"),
        (6, "日本語テスト"),
        (7, "اختبار عربي"),
    ])
    def test_string_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, value: str):
        """Test String roundtrip."""
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"str_{idx}", value)
        assert result == value

    @pytest.mark.parametrize("idx,value", [
        (0, b""),
        (1, bytes([0x00, 0x01, 0x02, 0xFF, 0xFE])),
        (2, bytes([0xDE, 0xAD, 0xBE, 0xEF])),
    ])
    def test_blob_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, value: bytes):
        """Test Blob roundtrip."""
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"blob_{idx}", value)
        assert result == value

    # --- Boolean Type ---

    @pytest.mark.parametrize("value", [True, False])
    def test_bool_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, value: bool):
        """Test Bool roundtrip."""
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"bool_{value}", value)
        assert result == value

    # --- Spatial Types ---

    @pytest.mark.parametrize("idx,lat,lon", [
        (0, 40.7128, -74.0060),
        (1, 0.0, 0.0),
        (2, 35.6762, 139.6503),
        (3, 90.0, 180.0),
        (4, -90.0, -180.0),
    ])
    def test_point_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, lat: float, lon: float):
        """Test Point roundtrip."""
        point = Point(latitude=lat, longitude=lon)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"point_{idx}", point)
        assert isinstance(result, Point)
        assert abs(result.latitude - lat) < 1e-10
        assert abs(result.longitude - lon) < 1e-10

    @pytest.mark.parametrize("idx,x,y,z", [
        (0, 0.0, 0.0, 0.0),
        (1, 1.0, 2.0, 3.0),
        (2, -1.5, -2.5, -3.5),
        (3, 1e10, 1e10, 1e10),
    ])
    def test_point3d_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, x: float, y: float, z: float):
        """Test Point3D roundtrip."""
        point = Point3D(x=x, y=y, z=z)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"point3d_{idx}", point)
        assert isinstance(result, Point3D)
        assert abs(result.x - x) < 1e-10
        assert abs(result.y - y) < 1e-10
        assert abs(result.z - z) < 1e-10

    # --- Special Types ---

    @pytest.mark.parametrize("idx,value", [
        (0, "123.456"),
        (1, "-987.654321"),
        (2, "12345"),
    ])
    def test_decimal_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, value: str):
        """Test Decimal roundtrip."""
        decimal = GqldbDecimal(value=value)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"decimal_{idx}", decimal)
        # Database may return decimal as string or GqldbDecimal object
        if isinstance(result, GqldbDecimal):
            assert result.value == value
        elif isinstance(result, str):
            assert result == value
        else:
            print(f"Decimal {idx} returned type: {type(result)}, value: {result}")
            assert result is not None

    @pytest.mark.parametrize("idx,values", [
        (0, [1.0, 2.0, 3.0, 4.0]),
        (1, [0.5, 0.5, 0.5, 0.5]),
        (2, [-1.0, -2.0, 0.0, 1.0]),
    ])
    def test_vector_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, values: list):
        """Test Vector roundtrip."""
        vector = Vector(values=values)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"vector_{idx}", vector)
        assert isinstance(result, Vector)
        assert len(result.values) == len(values)
        for i, v in enumerate(values):
            assert abs(result.values[i] - v) < 1e-5

    # --- Temporal Types ---

    @pytest.mark.parametrize("idx,nanos", [
        (0, 1718458245123456789),
        (1, 1718409600000000000),
    ])
    def test_local_datetime_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, nanos: int):
        """Test LocalDateTime roundtrip."""
        from datetime import datetime, timezone
        epoch_seconds = nanos // 1_000_000_000
        sub_nanos = nanos % 1_000_000_000
        dt = datetime.fromtimestamp(epoch_seconds, tz=timezone.utc)
        ldt = GqldbLocalDateTime(
            year=dt.year, month=dt.month, day=dt.day,
            hour=dt.hour, minute=dt.minute, second=dt.second,
            nanosecond=sub_nanos,
        )
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"ldt_{idx}", ldt)
        # Database may return as GqldbLocalDateTime or other type
        if isinstance(result, GqldbLocalDateTime):
            assert result.year == ldt.year
            assert result.month == ldt.month
            assert result.day == ldt.day
        else:
            print(f"LocalDateTime {idx} returned type: {type(result)}, value: {result}")
            assert result is not None

    @pytest.mark.skip(reason="Database may not support ZonedDateTime type in this schema")
    @pytest.mark.parametrize("idx,nanos,offset", [
        (0, 1718458245000000000, 0),
    ])
    def test_zoned_datetime_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, nanos: int, offset: int):
        """Test ZonedDateTime roundtrip."""
        zdt = GqldbZonedDateTime(nanos=nanos, offset_seconds=offset)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"zdt_{idx}", zdt)
        assert isinstance(result, GqldbZonedDateTime)

    @pytest.mark.parametrize("idx,hour,minute,second,nanosecond", [
        (0, 0, 0, 0, 0),
        (1, 12, 0, 0, 0),
    ])
    def test_local_time_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, hour: int, minute: int, second: int, nanosecond: int):
        """Test LocalTime roundtrip."""
        lt = GqldbLocalTime(hour=hour, minute=minute, second=second, nanosecond=nanosecond)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"lt_{idx}", lt)
        if isinstance(result, GqldbLocalTime):
            assert result.hour == hour and result.minute == minute and result.second == second
        else:
            print(f"LocalTime {idx} returned type: {type(result)}, value: {result}")
            assert result is not None

    @pytest.mark.skip(reason="Database may not support ZonedTime type in this schema")
    @pytest.mark.parametrize("idx,nanos,offset", [
        (0, 43200000000000, 0),
    ])
    def test_zoned_time_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, nanos: int, offset: int):
        """Test ZonedTime roundtrip."""
        zt = GqldbZonedTime(nanos_of_day=nanos, offset_seconds=offset)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"zt_{idx}", zt)
        assert isinstance(result, GqldbZonedTime)

    @pytest.mark.parametrize("idx,months", [
        (0, 0),
        (1, 12),
        (2, 18),
    ])
    def test_year_to_month_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, months: int):
        """Test YearToMonth roundtrip."""
        ytm = YearToMonth(months=months)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"ytm_{idx}", ytm)
        if isinstance(result, YearToMonth):
            assert result.months == months
        else:
            print(f"YearToMonth {idx} returned type: {type(result)}, value: {result}")
            assert result is not None

    @pytest.mark.skip(reason="Database may not support DayToSecond type in all graph schemas")
    @pytest.mark.parametrize("idx,nanos", [
        (0, 0),
        (1, 86400000000000),
    ])
    def test_day_to_second_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, nanos: int):
        """Test DayToSecond roundtrip."""
        dts = DayToSecond(seconds=nanos // 1_000_000_000, nanoseconds=nanos % 1_000_000_000)
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"dts_{idx}", dts)
        if isinstance(result, DayToSecond):
            assert result.seconds == nanos // 1_000_000_000
            assert result.nanoseconds == nanos % 1_000_000_000

    # --- Collection Types ---

    @pytest.mark.skip(reason="List type may not be supported in all graph schemas")
    @pytest.mark.parametrize("idx,value", [
        (0, ["a", "b", "c"]),
        (1, [1, 2, 3]),
    ])
    def test_list_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, value: list):
        """Test List roundtrip."""
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"list_{idx}", value)
        assert isinstance(result, list)
        assert len(result) == len(value)

    @pytest.mark.skip(reason="Map type may not be supported in all graph schemas")
    @pytest.mark.parametrize("idx,value", [
        (0, {"key1": "value1", "key2": 42}),
        (1, {"str": "text", "num": 123, "bool": True}),
    ])
    def test_map_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, value: dict):
        """Test Map roundtrip."""
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"map_{idx}", value)
        assert isinstance(result, dict)
        assert len(result) == len(value)

    @pytest.mark.skip(reason="Set type may not be supported in all graph schemas")
    @pytest.mark.parametrize("idx,value", [
        (0, {"a", "b", "c"}),
        (1, {1, 2, 3}),
    ])
    def test_set_roundtrip(self, connected_client: GqldbClient, roundtrip_graph: str, idx: int, value: set):
        """Test Set roundtrip."""
        result = self.insert_and_query_property(connected_client, roundtrip_graph, f"set_{idx}", value)
        # Set may be returned as list or set
        if isinstance(result, set):
            assert len(result) == len(value)
        else:
            assert len(result) == len(value)
